/*
*  E1.1.0 RAMS Increment 1 Patch Code
*  January 2015
*  Department of Veterans Affairs/Booz Allen Hamilton
*/

package gov.va.rams.service;

import gov.va.rams.docx4j.DocXFormGenerator;
import gov.va.rams.entities.ContainedForm;
import gov.va.rams.entities.Facility;
import gov.va.rams.entities.Form;
import gov.va.rams.entities.FormQuestion;
import gov.va.rams.entities.FormVersion;
import gov.va.rams.entities.QuestionResponse;
import gov.va.rams.entities.Study;
import gov.va.rams.entities.StudyStatus;
import gov.va.rams.entities.User;
import gov.va.rams.service.elements.FormValidationReport;
import gov.va.rams.service.elements.QuestionResponseUpdates;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.ejb.EJB;
import javax.servlet.ServletContext;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.persistence.EntityManager;
import javax.persistence.NoResultException;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;
import javax.ws.rs.Consumes;
import javax.ws.rs.GET;
import javax.ws.rs.PUT;
import javax.ws.rs.DELETE;
import javax.ws.rs.HeaderParam;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.ResponseBuilder;

/**
 *
 * @author Brian
 */
@Stateless
@Consumes(MediaType.APPLICATION_JSON)
@Produces(MediaType.APPLICATION_JSON)
@Path("")
public class ServiceREST {
    
    @javax.ws.rs.core.Context HttpHeaders requestHeaders;
    @javax.ws.rs.core.Context ServletContext context;

    @PersistenceContext(unitName = "RAMSPU")
    private EntityManager em;

    @GET
    @Path("forms")
    public List<FormVersion> getRootForms() {
        return em.createNamedQuery("FormVersion.findAll", FormVersion.class).getResultList();
    }
    
    @GET
    @Path("forms/{version}")
    public Form getRootForm(@PathParam("version") Integer version) {
        return em.find(FormVersion.class, version).getRootForm();
    }
    
    @GET
    @Path("user")
    public List<User> getUsers() {
        return em.createNamedQuery("User.findAll", User.class).getResultList();
    }

    @GET
    @Path("facility")
    public List<Facility> getFacilities() {
        return em.createNamedQuery("Facility.findAll", Facility.class).getResultList();
    }
    
    @GET
    @Path("study/{id}")
    public Study retrieveStudy(@PathParam("id") String id) {
        return em.createNamedQuery("Study.findByAssignedID", Study.class).setParameter("assignedID", id).getSingleResult();
    }
    
    @GET
    @Path("study/{id}/responses/{stage}")
    public Collection<QuestionResponse> retrieveStudyResponsesByStage(@PathParam("id") String id, @PathParam("stage") Integer stage) {
        Study s = em.createNamedQuery("Study.findByAssignedID", Study.class).setParameter("assignedID", id).getSingleResult();
        for (StudyStatus ss : s.getStudyStatusCollection())
            if (ss.getStageCounter() == stage)
                return ss.getQuestionResponseCollection();
        throw ServiceException.notFound();
    }

    @EJB
    private ResponseValidator responseValidator;
    
    @GET
    @Path("study/{id}/responses/{stage}/rollup")
    public QuestionResponseUpdates retrieveStudyResponsesRollup(@PathParam("id") String id, @PathParam("stage") Integer lastStage) {
        Study s = em.createNamedQuery("Study.findByAssignedID", Study.class).setParameter("assignedID", id).getSingleResult();
        List<QuestionResponse> responsesByStage = new ArrayList<>();
        QuestionResponseUpdates updates = new QuestionResponseUpdates();
        StudyStatus status = null;
        for (StudyStatus ss : s.getStudyStatusCollection()) {
            responsesByStage.addAll(ss.getQuestionResponseCollection());
            if (ss.getStageCounter() == lastStage) {
                updates.setUpdateTimestamp(ss.getUpdatedAt());
                status = ss;
                break;
            }
        }
        if (status == null)
            throw ServiceException.notFound();
        HashMap<QuestionResponse.Key, QuestionResponse> rollup = QuestionResponse.rollup(responsesByStage);
        updates.setResponses(rollup.values());
        updates.setFormValidationReport(responseValidator.getFormValidationReport(status));
        if (updates.getFormValidationReport() == null)
            updates.setFormValidationReport(responseValidator.updateFormValidationReport(status, rollup, false));
        return updates;
    }
    
    @GET
    @Path("study/{id}/responses/{stage}/validation")
    public FormValidationReport performStudyFormValidation(@PathParam("id") String id, @PathParam("stage") Integer lastStage) {
        Study s = em.createNamedQuery("Study.findByAssignedID", Study.class).setParameter("assignedID", id).getSingleResult();
        FormValidationReport report = null;
        for (StudyStatus ss : s.getStudyStatusCollection()) {
            if (ss.getStageCounter() == lastStage) {
                report = responseValidator.getFormValidationReport(ss);
                break;
            }
        }
        if (report == null) {
            List<QuestionResponse> responsesByStage = new ArrayList<>();
            StudyStatus status = null;
            for (StudyStatus ss : s.getStudyStatusCollection()) {
                responsesByStage.addAll(ss.getQuestionResponseCollection());
                if (ss.getStageCounter() == lastStage) {
                    status = ss;
                    break;
                }
            }
            if (status == null)
                throw ServiceException.notFound();
            report = responseValidator.updateFormValidationReport(status, QuestionResponse.rollup(responsesByStage), false);
        }
        return report;
    }

    @POST
    @Consumes
    @Path("study/{id}/responses/{stage}/cleanup")
    public void performResponseCleanup(@PathParam("id") String id, @PathParam("stage") Integer lastStage) {
        Study s = em.createNamedQuery("Study.findByAssignedID", Study.class).setParameter("assignedID", id).getSingleResult();
        StudyStatus latest = null;
        for (StudyStatus ss : s.getStudyStatusCollection()) {
            if (ss.getStageCounter() == lastStage)
                latest = ss;
            else if (ss.getStageCounter() > lastStage)
                throw ServiceException.forbidden();
        }
        if (latest == null)
            throw ServiceException.notFound();

        List<QuestionResponse> responsesByStage = new ArrayList<>();
        for (StudyStatus ss : s.getStudyStatusCollection())
            responsesByStage.addAll(ss.getQuestionResponseCollection());
        responseValidator.updateFormValidationReport(latest, QuestionResponse.rollup(responsesByStage), true);
    }
    
    @EJB
    private StudyStatusNotifier studyStatusNotifier;
    @EJB // We use this to separate out the transactional contexts into appropriate lifetimes.
    private QuestionResponseUpdater questionResponseUpdater;
    
    @POST
    @Path("study/{id}/responses/{stage}")
    @TransactionAttribute(TransactionAttributeType.NEVER)
    public void setStudyResponses(@PathParam("id") String id, @PathParam("stage") Integer stage, Collection<QuestionResponse> responses, @QueryParam("user") Integer usersID) {
        studyStatusNotifier.notifyUpdate(questionResponseUpdater.setStudyResponses(id, stage, responses, usersID));
    }
        
    @POST
    @Path("study/{id}/responses/{stage}/feed")
    public QuestionResponseUpdates retrieveStudyResponseFeed(@PathParam("id") String id, @PathParam("stage") Integer stage, QuestionResponseUpdates request, @QueryParam("form") Integer formsID, @HeaderParam("sm_user") String loginName) throws InterruptedException {
        Study s = em.createNamedQuery("Study.findByAssignedID", Study.class).setParameter("assignedID", id).getSingleResult();
        try {
            User u = em.createNamedQuery("User.findByLoginName", User.class).setParameter("loginName", loginName).getSingleResult();
            Form f = em.find(Form.class, formsID);
            studyFormLock.renew(s, f, u);
        } catch (NoResultException ignoreBadLocking) {}
        for (StudyStatus ss : s.getStudyStatusCollection()) {
            if (ss.getStageCounter() == stage) {
                QuestionResponseUpdates response = new QuestionResponseUpdates();
                response.setResponses(new LinkedList<QuestionResponse>());
                if (!ss.getUpdatedAt().after(request.getUpdateTimestamp())) {
                    // We haven't gotten updates, so let's sleep on it.
                    if (studyStatusNotifier.awaitUpdate(ss, request.getUpdateTimestamp()))
                        ss = questionResponseUpdater.getUpdate(ss);
                }
                response.setUpdateTimestamp(ss.getUpdatedAt());
                for (QuestionResponse qr : ss.getQuestionResponseCollection()) {
                    if (qr.getUpdatedAt().after(request.getUpdateTimestamp()))
                        response.getResponses().add(qr);
                }
                response.setFormValidationReport(responseValidator.getFormValidationReport(ss)); // Must be in the if it changed since the call to GET .../rollup!
                return response;
            }
        }
        throw ServiceException.notFound();
    }
    
    @EJB
    private StudyFormLock studyFormLock;
    
    @PUT
    @Path("study/{id}/lock/{formid}")
    public void acquireLock(@PathParam ("id") String studyId,@PathParam("formid") Integer formId,@HeaderParam("sm_user") String loginName) {
        Study s = em.createNamedQuery("Study.findByAssignedID", Study.class).setParameter("assignedID", studyId).getSingleResult();
        User u = em.createNamedQuery("User.findByLoginName",User.class).setParameter("loginName",loginName).getSingleResult();
        Form f = em.find(Form.class, formId);

        if (!studyFormLock.acquire(s,f,u))
            throw ServiceException.forbidden();
    }
    
    @DELETE
    @Path("study/{id}/lock/{formid}")
    public void releaseLock(@PathParam ("id") String studyId,@PathParam("formid") Integer formId,@HeaderParam("sm_user") String loginName) {
        Study s = em.createNamedQuery("Study.findByAssignedID", Study.class).setParameter("assignedID", studyId).getSingleResult();
        User u = em.createNamedQuery("User.findByLoginName",User.class).setParameter("loginName",loginName).getSingleResult();
        Form f = em.find(Form.class, formId);
        
        if (!studyFormLock.release(s,f,u))
            throw ServiceException.forbidden();
    }

    @DELETE
    @Path("study/{id}/lock_a/{formid}")
    public void adminReleaseLock(@PathParam ("id") String studyId,@PathParam("formid") Integer formId,@HeaderParam("sm_user") String loginName) {
        Study s = em.createNamedQuery("Study.findByAssignedID", Study.class).setParameter("assignedID", studyId).getSingleResult();
        User u = em.createNamedQuery("User.findByLoginName",User.class).setParameter("loginName",loginName).getSingleResult();
        Form f = em.find(Form.class, formId);
        
        System.out.println(s);
        System.out.println(u);
        System.out.println(f);
        
        if (studyFormLock.administrativeRelease(s,f,u)){
            System.out.println("\nLock Released\n");
        }else{
            System.out.println("\nLock Not Released\n");
            throw ServiceException.forbidden();
        }
    }

    //temporary for debugging
    @GET
    @Produces(MediaType.TEXT_PLAIN)
    @Path("queryLock")
    public String queryLock() {
        return studyFormLock.inspect();
    }
    
     
    @GET
    @Path("study/{id}/documentGeneration/{stage}/{which}") //   @Produces("application/docx")
    @Produces("application/docx")
    public Response generateDocument(  @PathParam("id") String id, @PathParam("stage") Integer lastStage, @PathParam("which") Integer sectionNumber) throws IOException {
    	System.out.println("service found");
    	System.out.println("Section number = " + sectionNumber);
    	
    	Study s = em.createNamedQuery("Study.findByAssignedID", Study.class).setParameter("assignedID", id).getSingleResult();
        Form root = em.find(FormVersion.class, s.getFormVersionsID()).getRootForm();
        Form mySection = null;
        int hipaaSectionNumber = -1;
        System.out.println("Root Form Narrative = " + root.getNarrative());
        System.out.println("Contained Form Collection Size = " + root.getContainedFormCollection().size());

        Collection<ContainedForm> cfCollection = root.getContainedFormCollection();
        Iterator<ContainedForm> cfCollectionIter = cfCollection.iterator();
        while(cfCollectionIter.hasNext()){
        	ContainedForm cf = cfCollectionIter.next();
        	System.out.println("Contain Form Forms Id = " + cf.getContainedFormsID() + ", sequence position = " + cf.getFormSequencePosition());
        	if(cf.getContainedForm().getTitle().toLowerCase().trim().contains("hipaa")){
            	hipaaSectionNumber = cf.getFormSequencePosition();
            	System.out.println("HIPAA Section Sequence number = " + hipaaSectionNumber);
            }
            if (cf.getFormSequencePosition() == sectionNumber) {
                mySection = cf.getContainedForm();
                System.out.println("Inside contained form loop - section found");
            }
            
        }
               
        if (mySection == null){
        	System.out.println("mySection is null");
        }
        
        List<QuestionResponse> responsesByStage = new ArrayList<>();
        StudyStatus status = null;
        System.out.println("Begin study status collection");
        for (StudyStatus ss : s.getStudyStatusCollection()) {
            responsesByStage.addAll(ss.getQuestionResponseCollection());
            if (ss.getStageCounter() == lastStage) {
                status = ss;
                System.out.println("Inside study status loop - study status found");
                break;
            }
        }
        if (status == null){
        	System.out.println("Status is null");
            throw ServiceException.notFound();
        }
        
        System.out.println("ArrayList: responsesByStage Size = " + responsesByStage.size());
        
        
        HashMap<QuestionResponse.Key, QuestionResponse> rollup = QuestionResponse.rollup(responsesByStage);
        System.out.println("Rollup size = " + rollup.size());
        
        //Stack<String> responses = new Stack<>();
        LinkedList<DocXFormGenerator.Response> responses = new LinkedList<>();
        
        System.out.println("Before Form Question Loop");
        System.out.println("Is mySection Null? : " + (mySection == null));
        
        
        System.out.println("Section Narrative = " + mySection.getNarrative());
        System.out.println("Section Title = " + mySection.getTitle());
        System.out.println("Section Form Collection size = " + mySection.getContainedFormCollection().size());
        System.out.println("Section Form Question Collection size = " + mySection.getFormQuestionCollection().size());
                
	    System.out.println("Outside form question loop - begin");
	    ArrayList<Integer> qrAnswersIDs = new ArrayList<Integer>();
	    HashMap<Integer, Integer> dependency = new HashMap<Integer, Integer>();
	    
	    for (FormQuestion q : mySection.getFormQuestionCollection()) {
	       	
			QuestionResponse qrtemp = new QuestionResponse();
			qrtemp.setQuestionsID(q.getQuestion().getQuestionsID());

			QuestionResponse qr = rollup.get(new QuestionResponse.Key(qrtemp));
			if(qr == null)
			{
				continue;
			}
			if(qr.getAnswersID() != null)
				qrAnswersIDs.add(qr.getAnswersID());
			
			if(q.getRequiredPredecessorAnswersID() != null){
				if(qr.getAnswersID() != null){
					dependency.put(qr.getAnswersID(), q.getRequiredPredecessorAnswersID());
				}
//				else if(qr.getAnswersID() == null){
//					dependency.put(q.getQuestion().getQuestionsID(), q.getRequiredPredecessorAnswersID());
//				}
			}
	    }
	    System.out.println("These are the answer id's");
	    System.out.println(qrAnswersIDs);
	    
	    
        for (FormQuestion q : mySection.getFormQuestionCollection()) {
        	       	
			QuestionResponse qrtemp = new QuestionResponse();
			qrtemp.setQuestionsID(q.getQuestion().getQuestionsID());

			QuestionResponse qr = rollup.get(new QuestionResponse.Key(qrtemp));
            
			
			if (q.getRequiredPredecessorAnswersID() == null){ //|| (!q.getRequiredPredecessorAnswersID().toString().trim().equals("")) )){ //don't want to answer questions that don't have predecessor
				//|| rollup.containsKey(q.getRequiredPredecessorAnswersID())
				//does not like (!q.getRequiredPredecessorAnswersID().toString().trim().equals("") which looks like it says if it is the equivalent of null
				//(q.getRequiredPredecessorAnswersID() != null) if it has a parent
				
				
				
				System.out.println("No parent!");
				if (qr == null){
//            	System.out.println("Question response is [null]");
            	//throw ServiceException.notFound();
                DocXFormGenerator.ResponseType type;
                final String name = q.getQuestion().getResponseFormat().getName();
                if (name.equals("radio"))
                    type = DocXFormGenerator.ResponseType.Boolean;
                else
                    type = DocXFormGenerator.ResponseType.String;
            	responses.add(new DocXFormGenerator.Response(type, null));
            	continue;
            }
           
            //System.out.println("Question ID = " + q.getQuestion().getQuestionsID() + " :  Question = " + q.getQuestion().getQuestionHyperText()+  " : User response = " + qr.getResponse());
     
            if (q.getQuestion().getResponseFormat().getName().equals("radio")) {
                if (qr.getAnswersID() == null){
                	//System.out.println("Question response answers is null");
                	responses.add(new DocXFormGenerator.Response(DocXFormGenerator.ResponseType.Boolean, null));
                	//throw ServiceException.notFound();
                } else {
                    //fix in to compare primatives    
                    boolean response = qr.getAnswersID().intValue() == q.getQuestion().getAnswerCollection().iterator().next().getAnswersID().intValue();
                    responses.add(new DocXFormGenerator.Response(DocXFormGenerator.ResponseType.Boolean, String.valueOf(response)));
                }
            } else { // string
                String response = qr.getResponse();
                /*if (response == null)
                    throw ServiceException.notFound();*/
                responses.add(new DocXFormGenerator.Response(DocXFormGenerator.ResponseType.String, response));
            }
        }
			else if(q.getRequiredPredecessorAnswersID() != null)
			{
				System.out.println("At question: " + q.getQuestion().getQuestionsID());
				//if(qrAnswersIDs.contains(q.getRequiredPredecessorAnswersID()))
				if(checkParent(q.getRequiredPredecessorAnswersID(),qrAnswersIDs, dependency))
				{
					if (q.getQuestion().getResponseFormat().getName().equals("radio")){
						
						boolean response = qr.getAnswersID().intValue() == q.getQuestion().getAnswerCollection().iterator().next().getAnswersID().intValue();   
						responses.add(new DocXFormGenerator.Response(DocXFormGenerator.ResponseType.Boolean, String.valueOf(response)));
					}
					else{
						String response = qr.getResponse();
						responses.add(new DocXFormGenerator.Response(DocXFormGenerator.ResponseType.String, response));
					}
				}
				else{//inserting place holders for the docx template at the 2nd level dependency
					if (q.getQuestion().getResponseFormat().getName().equals("radio")){
						responses.add(new DocXFormGenerator.Response(DocXFormGenerator.ResponseType.Boolean, null));
					}
					else{
						responses.add(new DocXFormGenerator.Response(DocXFormGenerator.ResponseType.String, " "));
					}
				}
			}
			else{//inserting place holders for the docx template
				if (q.getQuestion().getResponseFormat().getName().equals("radio")){
					responses.add(new DocXFormGenerator.Response(DocXFormGenerator.ResponseType.Boolean, null));
				}
				else{
					responses.add(new DocXFormGenerator.Response(DocXFormGenerator.ResponseType.String, " "));
				}
				//System.out.println("I have a parent!");
			}
        }
        
        String wordDocumentType = "InformedConsent";
        if (sectionNumber == hipaaSectionNumber) { //consent form
        	wordDocumentType = "HIPAA";
        }

        HashMap<Integer, Boolean> hm = new HashMap<Integer, Boolean>();
        ArrayList<Integer> qal = new ArrayList<Integer>();
 
   
        DocXFormGenerator docXFormGen = new DocXFormGenerator();
        //File file = new File("/usr/tomcat/cache/temp.docx");
        File file = new File("");
        try {
			file = docXFormGen.doGeneration(sectionNumber, responses, wordDocumentType); // , httpResponse.getOutputStream() 
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
               
        ResponseBuilder response = Response.ok((Object) file);
        response.header("Content-Disposition", "attachment; filename="+file.getName());
        return response.build();
 
    }

    public boolean checkParent(int parentAnswerID, ArrayList<Integer> qrAnswersIDs, HashMap<Integer,Integer> dependency){
    	//module for checking 2nd level dependencies
    	System.out.println("checking parent: " + parentAnswerID);
    	
    	if(!qrAnswersIDs.contains(parentAnswerID)){//if you are not in the rollup
    		//System.out.println("parent " + parentAnswerID + " isn't checked");
    		return false;
    	}else if(dependency.containsKey(parentAnswerID)){//if you have a parent who has a grandparent
    		//System.out.println("parent + " + parentAnswerID + " is checked and has a grandparent + " + dependency.get(parentAnswerID));
    		return checkParent(dependency.get(parentAnswerID), qrAnswersIDs, dependency);
    	}
    	
    	//System.out.println("parent " + parentAnswerID + " is checked");
    	return true;
    }

}
