/*
* E1.0.0 RAM Increment 1 Final Code
* December 2014
* Department of Veterans Affairs/Booz Allen Hamilton 
* 
*/

package gov.va.rams.service;

import gov.va.rams.entities.QuestionResponse;
import gov.va.rams.entities.Study;
import gov.va.rams.entities.StudyStatus;
import gov.va.rams.entities.User;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;

/**
 * Setting and retrieving updated responses for a given StudyStatus is performed
 * via this EJB to guarantee that they occur in isolated JPA transactions and not
 * in a transaction that covers the entire REST service call.  If we did not do
 * this transaction isolation, we could notify asynchronous update listeners
 * that there is an update before it has been committed to the database, and we
 * could get stale data after waiting for updates and receiving a notification
 * as the database continues to give us a consistent view for the duration of
 * the call in the context of {@link gov.va.rams.service.ServiceREST#retrieveStudyResponseFeed(java.lang.String, java.lang.Integer, gov.va.rams.service.elements.QuestionResponseUpdates) }.
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
public class QuestionResponseUpdater {
    @PersistenceContext(unitName = "RAMSPU")
    private EntityManager em;
    
    public StudyStatus getUpdate(StudyStatus fromOldTxact) {
        StudyStatus ss = em.find(StudyStatus.class, fromOldTxact.getStudyStatusID());
        ss.getQuestionResponseCollection();
        return ss;
    }
    
    @EJB
    private ResponseValidator responseValidator;

    public StudyStatus setStudyResponses(String id, Integer stage, Collection<QuestionResponse> responses, Integer usersID) {
        Study s = em.createNamedQuery("Study.findByAssignedID", Study.class).setParameter("assignedID", id).getSingleResult();
        final User user = usersID != null ? em.find(User.class, usersID) : null;
        List<QuestionResponse> responsesByStage = new ArrayList<>();
        int onlyModifyCurrent = s.getStudyStatusCollection().size();
        for (StudyStatus ss : s.getStudyStatusCollection()) {
            onlyModifyCurrent--;
            if (ss.getStageCounter() < stage) {
                responsesByStage.addAll(ss.getQuestionResponseCollection());
            } else {
                if (onlyModifyCurrent != 0)
                    throw ServiceException.forbidden();
                final Date timestamp = new Date();
                for (QuestionResponse newResponse : responses) {
                    // Does a response already exist?
                    QuestionResponse found = null;
                    final QuestionResponse.Key searchKey = new QuestionResponse.Key(newResponse);
                    for (QuestionResponse existingResponse : ss.getQuestionResponseCollection()) {
                        if (new QuestionResponse.Key(existingResponse).equals(searchKey)) {
                            found = existingResponse;
                            break;
                        }
                    }
                    if (found != null) {
                        found.setAnswersID(newResponse.getAnswersID());
                        found.setResponse(newResponse.getResponse());
                        found.setUpdatedAt(timestamp);
                    } else {
                        newResponse.setUpdatedAt(timestamp);
                        newResponse.setStudyStatus(ss);
                        ss.getQuestionResponseCollection().add(newResponse);
                    }
                }
                ss.setUpdatedAt(timestamp);
                if (user != null)
                    ss.setUpdatedBy(user);
                responsesByStage.addAll(ss.getQuestionResponseCollection());
                responseValidator.updateFormValidationReport(ss, QuestionResponse.rollup(responsesByStage));
                return ss;
            }
        }
        throw ServiceException.notFound();
    }
}
